# Copyright 2019 Dragonchain, Inc. or its affiliates. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class DragonchainException(Exception):
    """Base class exception"""


class EmptyUpdateException(DragonchainException):
    """Raised when user tries to update a resource with nothing to update"""


class AuthorizationNotFound(DragonchainException):
    """Raised when no form of authorization can be found"""


class DragonchainIdentityNotFound(DragonchainException):
    """Raised when no form of dragonchain ID can be found"""


class DragonchainServiceException(DragonchainException):
    """Raised when the request to the Dragonchain service was fulfilled, but responded with a non 200 status code"""


class ConnectionException(DragonchainException):
    """Raised when the request to the Dragonchain service was fulfilled, but responded with a non 200 status code"""


class UnexpectedResponseException(DragonchainException):
    """Raised when the Dragonchain responded with an unexpected response"""



# Copyright 2019 Dragonchain, Inc. or its affiliates. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
from dragonchain_sdk.request import Request
from dragonchain_sdk.credentials import Credentials

logger = logging.getLogger(__name__)


class Client(object):
    """Construct a new ``Client`` object

    Args:
        dragonchain_id (str): The ID of the chain to connect to.
        auth_key_id (str, optional): The authorization key ID
        auth_key (str, optional): The authorization key
        endpoint (str, optional): The endpoint of the Dragonchain
        verify (bool, optional): Verify the TLS cert of the Dragonchain
        algorithm (str, optional): The hashing algorithm used for HMAC authentication

    Returns:
        A new Dragonchain client.
    """
    def __init__(self, dragonchain_id=None, auth_key_id=None, auth_key=None, endpoint=None, verify=True, algorithm='SHA256'):
        self.credentials = Credentials(dragonchain_id, auth_key, auth_key_id, algorithm)
        self.request = Request(self.credentials, endpoint, verify)
        logger.debug('Client finished initialization')

    def override_credentials(self, auth_key=None, auth_key_id=None, dragonchain_id=None, update_endpoint=True):
        """Change dragonchain_id, auth_key or auth_key_id for this client

        Args:
            auth_key (str, optional): The authorization key
            auth_key_id (str, optional): The authorization key ID
            dragonchain_id (str, optional): The Dragonchain ID
            update_endpoint (bool, optional): If endpoint should automatically be updated for a new dragonchain_id

        Returns:
            None, sets instance variables on credentials service
        """
        if auth_key is not None:
            if isinstance(auth_key, str):
                self.credentials.auth_key = auth_key
            else:
                raise TypeError('Parameter "auth_key" must be of type str.')
        if auth_key_id is not None:
            if isinstance(auth_key, str):
                self.credentials.auth_key_id = auth_key_id
            else:
                raise TypeError('Parameter "auth_key_id" must be of type str.')
        if dragonchain_id is not None:
            if isinstance(dragonchain_id, str):
                self.credentials.dragonchain_id = dragonchain_id
                if update_endpoint is True:
                    self.request.update_endpoint()
            else:
                raise TypeError('Parameter "dragonchain_id" must be of type str.')

        logger.info('Credentials updated')

    def get_secret(self, secret_name):
        """Gets secrets for smart contracts

        Args:
            secret_name (str): name of the secret to retrieve
        Returns:
            Returns the secret from the file location
        """
        if not isinstance(secret_name, str):
            raise TypeError('Parameter "secret_name" must be of type str.')
        if not os.environ.get('SMART_CONTRACT_ID'):
            raise RuntimeError('Missing "SMART_CONTRACT_ID" from environment')
        path = os.path.join(os.path.abspath(os.sep), 'var', 'openfaas', 'secrets', 'sc-{}-{}'.format(os.environ.get('SMART_CONTRACT_ID'), secret_name))
        return open(path, 'r').read()

    def get_status(self):
        """Make a PUT request to a chain

        Returns:
            Returns the status of a Dragonchain
        """
        return self.request.get('/status')

    def query_contracts(self, query=None, sort=None, offset=0, limit=10):
        """Perform a query on a chain's smart contracts

        Args:
            query (str, optional): Lucene query parameter (e.g.: ``is_serial:true``)
            sort (str, optional): Sort syntax of 'field:direction' (e.g.: ``name:asc``)
            offset (int, optional): Pagination offset of query (default 0)
            limit (int, optional): Pagination limit (default 10)

        Returns:
            The results of the query
        """
        query_params = self.request.get_lucene_query_params(query, sort, offset, limit)
        return self.request.get('/contract{}'.format(query_params))

    def get_contract(self, contract_id=None, txn_type=None):
        """Perform a query on a chain's smart contracts

        Args:
            contract_id (str, exclusive): Id of the contract to get
            txn_type (str, exclusive): Name of the contract to get

        Returns:
            The contract returned from the request
        """
        if contract_id is not None and not isinstance(contract_id, str):
            raise TypeError('Parameter "contract_id" must be of type str.')
        if txn_type is not None and not isinstance(txn_type, str):
            raise TypeError('Parameter "txn_type" must be of type str.')
        if contract_id and txn_type:
            raise TypeError('Only one of "contract_id" and "txn_type" can be specified')
        if contract_id:
            return self.request.get('/contract/{}'.format(contract_id))
        elif txn_type:
            return self.request.get('/contract/txn_type/{}'.format(txn_type))
        else:
            raise TypeError('At least one of "contract_id" or "txn_type" must be specified')

    def post_contract(self, txn_type, execution_order, image, cmd, args=None, env=None, secrets=None, seconds=None, cron=None, auth=None):
        """Post a contract to a chain

        Args:
            txn_type (str): txn_type of the contract to create
            execution_order (string): Order of execution. Valid values are 'serial' or 'parallel'
            image (str): Docker image containing the smart contract logic
            cmd (str): Entrypoint command to run in the docker container
            args (list, optional): List of arguments to the cmd field
            env (dict, optional): dict mapping of environment variables for your contract runtime
            secrets (dict, optional): dict mapping of secrets for your contract runtime
            seconds (string, optional): The seconds of scheduled execution in seconds
            cron (string, optional): The rate of scheduled execution specified as a cron
            auth (string, optional): basic-auth for pulling docker images, base64 encoded (e.g. username:password)

        Returns:
            Success or failure object
        """
        if not isinstance(txn_type, str):
            raise TypeError('Parameter "txn_type" must be of type str.')
        if not isinstance(image, str):
            raise TypeError('Parameter "image" must be of type str.')
        if not isinstance(cmd, str):
            raise TypeError('Parameter "cmd" must be of type str.')
        if not isinstance(execution_order, str):
            raise TypeError('Parameter "execution_order" must be of type str.')
        if args is not None and not isinstance(args, list):
            raise TypeError('Parameter "args" must be of type list.')
        if env is not None and not isinstance(env, dict):
            raise TypeError('Parameter "env" must be of type dict.')
        if secrets is not None and not isinstance(secrets, dict):
            raise TypeError('Parameter "secrets" must be of type dict.')
        if seconds is not None and not isinstance(seconds, int):
            raise TypeError('Parameter "seconds" must be of type int.')
        if cron is not None and not isinstance(cron, str):
            raise TypeError('Parameter "cron" must be of type str.')
        if auth is not None and not isinstance(auth, str):
            raise TypeError('Parameter "auth" must be of type str.')

        body = {
            'version': '3',
            'txn_type': txn_type,
            'image': image,
            'cmd': cmd,
            'execution_order': execution_order
        }
        if env:
            body['env'] = env
        if args:
            body['args'] = args
        if secrets:
            body['secrets'] = secrets
        if seconds:
            body['seconds'] = seconds
        if cron:
            body['cron'] = cron
        if auth:
            body['auth'] = auth
        return self.request.post('/contract', body)

    def update_contract(self, contract_id, execution_order=None, image=None, cmd=None, args=None, desired_state=None, env={}, secrets={}, seconds=None, cron=None, auth=None):
        """Update an existing smart contract. The contract_id and at least one optional parameter must be supplied.

        Args:
            contract_id (str): contract_id of the contract to create
            execution_order (string): Order of execution. Valid values are 'serial' or 'parallel'
            image (str): Docker image containing the smart contract logic
            cmd (str): Entrypoint command to run in the docker container
            args (list, optional): List of arguments to the cmd field
            desired_state (str, optional): Change the state of a contract. Valid values are "active" and "inactive". You may only change the state of an active or inactive contract.
            env (dict, optional): dict mapping of environment variables for your contract runtime
            secrets (dict, optional): dict mapping of secrets for your contract runtime
            seconds (int, optional): The seconds of scheduled execution in seconds
            cron (string, optional): The rate of scheduled execution specified as a cron
            auth (string, optional): basic-auth for pulling docker images, base64 encoded (e.g. username:password)

        Returns:
            Success or failure object
        """
        if not isinstance(contract_id, str):
            raise TypeError('Parameter "contract_id" must be of type str.')
        if image is not None and not isinstance(image, str):
            raise TypeError('Parameter "image" must be of type str.')
        if cmd is not None and not isinstance(cmd, str):
            raise TypeError('Parameter "cmd" must be of type str.')
        if execution_order is not None and not isinstance(execution_order, str):
            raise TypeError('Parameter "execution_order" must be of type str.')
        if desired_state is not None and not isinstance(desired_state, str):
            raise TypeError('Parameter "desired_state" must be of type str.')
        if args is not None and not isinstance(args, list):
            raise TypeError('Parameter "args" must be of type list.')
        if env is not None and not isinstance(env, dict):
            raise TypeError('Parameter "env" must be of type dict.')
        if secrets is not None and not isinstance(secrets, dict):
            raise TypeError('Parameter "secrets" must be of type dict.')
        if seconds is not None and not isinstance(seconds, int):
            raise TypeError('Parameter "seconds" must be of type int.')
        if cron is not None and not isinstance(cron, str):
            raise TypeError('Parameter "cron" must be of type str.')
        if auth is not None and not isinstance(auth, str):
            raise TypeError('Parameter "auth" must be of type str.')

        body = {
            'version': '3'
        }
        if image:
            body['image'] = image
        if cmd:
            body['cmd'] = cmd
        if execution_order:
            body['execution_order'] = execution_order
        if desired_state:
            body['desired_state'] = desired_state
        if args:
            body['args'] = args
        if env:
            body['env'] = env
        if secrets:
            body['secrets'] = secrets
        if seconds:
            body['seconds'] = seconds
        if cron:
            body['cron'] = cron
        if auth:
            body['auth'] = auth

        return self.request.put('/contract/{}'.format(contract_id), body)

    def delete_contract(self, contract_id):
        """Delete an existing contract

        Args:
            contract_id (str): Transaction type of the contract to delete

        Returns:
            The results of the delete request
        """
        if not isinstance(contract_id, str):
            raise TypeError('Parameter "state" must be of type str.')
        return self.request.delete('/contract/{}'.format(contract_id), body={})

    def query_transactions(self, query=None, sort=None, offset=0, limit=10):
        """Perform a query on a chain's transactions

        Args:
            query (str, optional): Lucene query parameter (e.g.: ``is_serial:true``)
            sort (str, optional): Sort syntax of 'field:direction' (e.g.: ``txn_type:asc``)
            offset (int, optional): Pagination offset of query (default 0)
            limit (int, optional): Pagination limit (default 10)

        Returns:
            The results of the query
        """
        query_params = self.request.get_lucene_query_params(query, sort, offset, limit)
        return self.request.get('/transaction{}'.format(query_params))

    def get_transaction(self, txn_id):
        """Get a specific transaction by id

        Args:
            txn_id (str, uuid): ID of the transaction to get

        Returns:
            The transaction searched for
        """
        if not isinstance(txn_id, str):
            raise TypeError('Paramter "txn_id" must be of type str.')
        return self.request.get('/transaction/{}'.format(txn_id))

    def post_transaction(self, txn_type, payload, tag=None):
        """Post a transaction to a chain

        Args:
            txn_type (str): Type of transaction
            payload (dict or string): The payload of the transaction
            tag (str, optional): A tag string to search on

        Returns:
            Transaction ID on success
        """
        return self.request.post('/transaction', Client.build_transaction_dict(txn_type, payload, tag))

    def post_transaction_bulk(self, txn_list):
        """Post many transactions to a chain at once, over a single connnection

        Args:
            txn_list (list): List of transaction dictionaries. Schema: ``{'txn_type': 'str', 'payload': 'str or dict', 'tag': 'str (optional)'}``

        Returns:
            List of succeeded transaction id's and list of failed transactions
        """
        if not isinstance(txn_list, list):
            raise TypeError('Parameter "txn_list" must be of type list.')

        post_data = []

        for txn in txn_list:
            if not isinstance(txn, dict):
                raise TypeError('All items in parameter "txn_list" must be of type dict.')
            post_data.append(Client.build_transaction_dict(txn.get('txn_type'), txn.get('payload'), txn.get('tag')))

        return self.request.post('/transaction_bulk', post_data)

    @staticmethod
    def build_transaction_dict(txn_type, payload, tag=None):
        if not isinstance(txn_type, str):
            raise TypeError('Parameter "txn_type" must be of type str.')
        if not isinstance(payload, str) and not isinstance(payload, dict):
            raise TypeError('Parameter "payload" must be of type dict or str.')
        if tag is not None and not isinstance(tag, str):
            raise TypeError('Parameter "tag" must be of type str.')

        body = {
            'version': '1',
            'txn_type': txn_type,
            'payload': payload
        }
        if tag:
            body['tag'] = tag

        return body

    def query_blocks(self, query=None, sort=None, offset=0, limit=10):
        """Perform a query on a chain's blocks

        Args:
            query (str, optional): Lucene query parameter (e.g.: ``is_serial:true``)
            sort (str, optional): Sort syntax of 'field:direction' (e.g.: ``block_id:asc``)
            offset (int, optional): Pagination offset of query (default 0)
            limit (int, optional): Pagination limit (default 10)

        Returns:
            The results of the query
        """
        query_params = self.request.get_lucene_query_params(query, sort, offset, limit)
        return self.request.get('/block{}'.format(query_params))

    def get_block(self, block_id):
        """Get a specific block by id

        Args:
            block_id (int): ID of the block to get

        Returns:
            The block searched for
        """
        if not isinstance(block_id, str) and not isinstance(block_id, int):
            raise TypeError('Parameter "block_id" must be of type str or int.')
        return self.request.get('/block/{}'.format(block_id))

    def get_verification(self, block_id, level=None):
        """Get higher level block verifications by level 1 block id

        Args:
            block_id (int): ID of the block to get
            level (int): Level of verifications to get (valid values are 2, 3, 4 and 5)

        Returns:
            Higher level block verifications
        """
        if not isinstance(block_id, str) and not isinstance(block_id, int):
            raise TypeError('Parameter "block_id" must be of type str.')
        if level is not None:
            if not isinstance(level, int) and not isinstance(level, str):
                raise TypeError('Parameter "level" must be of type str or int.')
            if int(level) not in [2, 3, 4, 5]:
                raise ValueError('Parameter "level" must be between 2 and 5 inclusive.')
            return self.request.get('/verifications/{}?level={}'.format(block_id, level))
        return self.request.get('/verifications/{}'.format(block_id))

    def get_sc_heap(self, sc_id=None, key=None):
        """Retrieve data from the heap storage of a smart contract
        Note: When ran in an actual smart contract, sc_id will be pulled automatically from the environment if not explicitly provided

        Args:
            key (str): The key stored in the heap to retrieve
            sc_id (str, optional): The ID of the smart contract, optional if called from within a smart contract

        Returns:
            The value of the object in the heap
        """
        if not isinstance(key, str):
            raise TypeError('Parameter "key" must be of type str.')
        if sc_id is None:
            sc_id = os.environ.get('SMART_CONTRACT_ID')
        if not isinstance(sc_id, str):
            raise TypeError('Parameter "sc_id" must be of type str.')
        return self.request.get('/get/{}/{}'.format(sc_id, key), parse_response=False)

    def list_sc_heap(self, sc_id=None, folder=None):
        """Lists all objects stored in a smart contracts heap
        Note: When ran in an actual smart contract, sc_id will be pulled automatically from the environment if not explicitly provided

        Args:
            sc_id (str, optional): sc_id heap to list. If not provided explicitly, it must be in the SMART_CONTRACT_NAME env var
            folder (str, optional): the folder to list in the heap. If not provided, it will default to the root of the heap

        Returns:
            Parsed json response from the chain
        """
        if sc_id is None:
            sc_id = os.environ.get('SMART_CONTRACT_ID')
        if not isinstance(sc_id, str):
            raise TypeError('Parameter "sc_id" must be of type str.')
        if folder is not None:
            if not isinstance(folder, str):
                raise TypeError('Parameter "folder" must be of type str.')
            if folder.endswith('/'):
                raise ValueError('Parameter "folder" cannot end with /.')
            return self.request.get('/list/{}/{}/'.format(sc_id, folder))
        return self.request.get('/list/{}/'.format(sc_id))

    def get_transaction_type(self, transaction_type):
        """Gets information on a registered transaction type

        Args:
            transaction_type (str): transaction_type to retrieve data for

        Returns:
            parsed json response of the transaction type or None
        """
        if not isinstance(transaction_type, str):
            raise TypeError('Parameter "transaction_type" must be of type str.')
        return self.request.get('/transaction-type/{}'.format(transaction_type))

    def list_transaction_types(self):
        """Lists out all registered transaction types for a chain

        Returns:
            list of registered transaction types
        """
        return self.request.get('/transaction-types')

    def update_transaction_type(self, transaction_type, custom_indexes=None):
        """Updates the custom index of a given registered transaction type

        Args:
            transaction_type (str): transaction_type to update
            custom_indexes (list): custom_indexes to update

        Returns:
            Parsed json with success message
        """
        if custom_indexes is None:
            custom_indexes = []
        if not isinstance(transaction_type, str):
            raise TypeError('Parameter "transaction_type" must be of type str.')
        if not isinstance(custom_indexes, list):
            raise TypeError('Parameter "custom_indexes" must be of type list.')
        params = {"version": "1", "custom_indexes": custom_indexes}
        return self.request.put('/transaction-type/{}'.format(transaction_type), params)

    def register_transaction_type(self, transaction_type, custom_indexes=None):
        """Registers a new custom index

        Args:
            transaction_type (str): transaction_type to update
            custom_indexes (list): custom_indexes to update

        Returns:
            Parsed json with success message
        """
        if not isinstance(transaction_type, str):
            raise TypeError('Parameter "transaction_type" must be of type str.')
        if custom_indexes and not isinstance(custom_indexes, list):
            raise TypeError('Parameter "custom_indexes" must be of type list.')
        params = {"version": "1", "txn_type": transaction_type}
        if custom_indexes:
            params['custom_indexes'] = custom_indexes
        return self.request.post('/transaction-type', params)

    def delete_transaction_type(self, transaction_type):
        """Deletes a transaction type registration

        Args:
            transaction_type (str): transaction_type to delete

        Returns:
            Parsed json with success message
        """
        if not isinstance(transaction_type, str):
            raise TypeError('Parameter "transaction_type" must be of type str.')
        return self.request.delete('/transaction-type/{}'.format(transaction_type), body={})



# Copyright 2019 Dragonchain, Inc. or its affiliates. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import requests
import logging
import urllib
from json import dumps
from dragonchain_sdk.credentials import Credentials
from dragonchain_sdk import exceptions

logger = logging.getLogger(__name__)

supported_http = {
    'GET': requests.get,
    'POST': requests.post,
    'PUT': requests.put,
    'PATCH': requests.patch,
    'DELETE': requests.delete,
    'HEAD': requests.head,
    'OPTIONS': requests.options
}


class Request(object):
    """Construct a new `Request` object

    Args:
        credentials (Credentials): The credentials for the chain to associate with requests
        endpoint (str, optional): The URL for the endpoint of the chain
        verify (bool, optional): Boolean indicating whether to validate the SSL certificate of the endpoint when making requests

    Returns:
        A new Request object.
    """
    def __init__(self, credentials, endpoint=None, verify=True):
        if isinstance(credentials, Credentials):
            self.credentials = credentials
        else:
            raise TypeError('Parameter "credentials" must be of type Credentials.')

        if isinstance(verify, bool):
            self.verify = verify
            logger.debug('SSL certificates {} be verified when making requests'.format('will' if self.verify else 'will NOT'))
        else:
            raise TypeError('Parameter "verify" must be of type bool.')

        self.update_endpoint(endpoint)

    def update_endpoint(self, endpoint=None):
        """Update endpoint for this request object

        Args:
            endpoint (str, optional): Endpoint to set. Will auto-generate based on credentials dragonchain_id if not provided

        Returns:
            None, sets the endpoint of this Request instance
        """
        if endpoint is None:
            self.endpoint = 'https://{}.api.dragonchain.com'.format(self.credentials.dragonchain_id)
        elif isinstance(endpoint, str):
            self.endpoint = endpoint
        else:
            raise TypeError('Parameter "endpoint" must be of type str.')
        logger.info('Target endpoint updated to {}'.format(self.endpoint))

    def get(self, path, parse_response=True):
        """Make a GET request to a chain

        Args:
            path (str): Path of the request (including any path query parameters)
            parse_response (bool, optional): Decides whether the return from the chain should be parsed as json (default True)

        Returns:
            Returns the response of the GET operation.
        """
        return self._make_request(http_verb='GET',
                                  path=path,
                                  verify=self.verify,
                                  parse_response=parse_response)

    def post(self, path, body, parse_response=True):
        """Make a POST request to a chain

        Args:
            path (str): Path of the request (including any path query parameters)
            body (dict): A dictionary representing the JSON to post.
            parse_response (bool, optional): Decides whether the return from the chain should be parsed as json (default True)

        Returns:
            Returns the response of the POST operation.
        """
        return self._make_request(http_verb='POST',
                                  path=path,
                                  verify=self.verify,
                                  json=body,
                                  parse_response=parse_response)

    def put(self, path, body, parse_response=True):
        """Make a PUT request to a chain

        Args:
            path (str): Path of the request (including any path query parameters)
            body (dict): A dictionary representing the JSON to put.
            parse_response (bool, optional): Decides whether the return from the chain should be parsed as json (default True)

        Returns:
            Returns the response of the PUT operation.
        """
        return self._make_request(http_verb='PUT',
                                  path=path,
                                  verify=self.verify,
                                  json=body,
                                  parse_response=parse_response)

    def delete(self, path, body, parsed_response=True):
        """Make a DELETE request to the chain

        Args:
            path (str): Path of the request (including any path query parameters)
            body (dict): A dictionary representing the JSON to put.
            parse_response (bool, optional): Decides whether the return from the chain should be parsed as json (default True)

        Returns:
            Returns the response of the DELETE operation
        """
        return self._make_request(http_verb='DELETE',
                                  path=path,
                                  verify=self.verify,
                                  json=body,
                                  parse_response=parsed_response)

    def get_requests_method(self, http_verb):
        """Get the appropriate requests method for a given http_verb

        Args:
            http_verb (str): the type of http request to make (GET, POST, etc)

        Returns:
            appropriate requests http method
        """
        if not isinstance(http_verb, str):
            raise TypeError('Parameter "http_verb" must be of type str.')
        request_method = supported_http.get(http_verb.upper())
        if not request_method:
            raise ValueError(http_verb + ' is an unsupported http operation.')
        return request_method

    def generate_query_string(self, query_dict):
        """Generate an http query string from a dictionary

        Args:
            query_dict (dict): dict of parameters to use in the query string

        Returns:
            query string to use in an HTTP request path
        """
        if not isinstance(query_dict, dict):
            raise TypeError('Parameter "query_dict" must be of type dict.')
        if query_dict:
            query = '?'
            query_string = urllib.parse.urlencode(query_dict)
            logger.debug('Generated query string {}'.format(query_string))
            return '{}{}'.format(query, query_string)
        else:
            # If input is empty, return an empty string as the query string
            return ''

    def get_lucene_query_params(self, query=None, sort=None, offset=0, limit=10):
        """Generate a lucene query param string with given inputs

        Args:
            query (str): lucene query parameter (e.g.: is_serial:true)
            sort (str): sort syntax of 'field:direction' (e.g.: name:asc)
            offset (int): pagination offset of query
            limit (int): pagination limit

        Returns:
            Query string to include in request path
        """
        if query is not None and not isinstance(query, str):
            raise TypeError('Parameter "query" must be of type str.')
        if sort is not None and not isinstance(sort, str):
            raise TypeError('Parameter "sort" must be of type str.')
        if not isinstance(offset, int) or not isinstance(limit, int):
            raise TypeError('Parameters "limit" and "offset" must be of type int.')
        params = {
            'offset': offset,
            'limit': limit
        }
        if query:
            params['q'] = query
        if sort:
            params['sort'] = sort
        return self.generate_query_string(params)

    def make_headers(self, timestamp, authorization, content_type=None):
        """Create a headers dictionary to send with a request to a dragonchain

        Args:
            timestamp (str): unix timestamp to put for this request
            content_type (str): content_type to use for this request
            authorization (str): authorization header to include with the request

        Returns:
            dict of headers to use with the request
        """
        if not isinstance(timestamp, str):
            raise TypeError('Parameter "timestamp" must be of type str.')
        if not isinstance(authorization, str):
            raise TypeError('Parameter "authorization" must be of type str.')
        if content_type is not None and not isinstance(content_type, str):
            raise TypeError('Parameter "content_type" must be of type str.')
        header_dict = {
            'dragonchain': self.credentials.dragonchain_id,
            'timestamp': timestamp,
            'Authorization': authorization
        }
        if content_type:
            header_dict['Content-Type'] = content_type
        return header_dict

    def _make_request(self, http_verb, path, json=None, timeout=30, verify=True, parse_response=True):
        """Make an http request to a dragonchain with the given information

        Args:
            http_verb (str): the type of http request to make (GET, POST, etc)
            path (str): the full path to make the request (including query params if any) starting with a '/'
            json (dict, optional): dictionary object to send as json (automatically sets content-type to application/json)
            timeout (int, optional): the timeout to wait for the dragonchain to respond (defaults to 30 seconds if not set)
            verify (bool, optional): specify if the SSL cert of the chain should be verified
            parse_response (bool, optional): if the return from the chain should be parsed as json

        Returns:
            Dictionary where status is the HTTP status code, response is the return body from the chain, and ok is a boolean if the status code was in the 2XX range
            {
                'status': int (http response code from chain)
                'ok': boolean (if http response code is 200-299)
                'response': dict if parse_response, else str (actual response body from chain)
            }
        """
        if not isinstance(path, str):
            raise TypeError('Parameter "path" must be of type str.')
        if not path.startswith('/'):
            raise ValueError('Parameter "path" must start with a \'/\'.')

        logger.debug('Creating request {} {}'.format(http_verb, path))
        requests_method = self.get_requests_method(http_verb)
        content_type = 'application/json'
        content = dumps(json)
        # Add the 'Z' manually to indicate UTC (not added by isoformat)
        timestamp = datetime.datetime.utcnow().isoformat() + 'Z'
        authorization = self.credentials.get_authorization(http_verb, path, timestamp, content_type, content)
        header_dict = self.make_headers(timestamp, authorization, content_type)
        full_url = self.endpoint + path

        logger.debug('Making request. Verify SSL: {}, Timeout: {}'.format(verify, timeout))
        logger.debug('{} {}'.format(http_verb, full_url))
        logger.debug('Headers: {}'.format(header_dict))
        logger.debug('Data: {}'.format(content))

        # Make request with appropriate data
        try:
            r = requests_method(url=full_url, data=content, headers=header_dict, timeout=timeout, verify=verify)
        except Exception as e:
            raise exceptions.ConnectionException('Error while communicating with the Dragonchain: {}'.format(e))
        return_dict = {}
        # Generate the return dictionary
        try:
            return_dict['status'] = r.status_code
            logger.debug('Response status code: {}'.format(r.status_code))
            return_dict['ok'] = True if r.status_code // 100 == 2 else False
            return_dict['response'] = r.json() if parse_response else r.text
            return return_dict
        except Exception as e:
            raise exceptions.UnexpectedResponseException('Unexpected response from Dragonchain. Response: {} | Error: {}'.format(r.text, e))



# Copyright 2019 Dragonchain, Inc. or its affiliates. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import configparser
import hashlib
import hmac
import os
import sys
import base64
import logging
from dragonchain_sdk import exceptions

logger = logging.getLogger(__name__)


class Credentials(object):
    """Construct a new `Credentials` object

    Args:
        dragonchain_id (str, optional): The dragonchain ID for these credentials
        auth_key (str, optional): The auth_key to use with these credentials
        auth_key_id (str, optional): The auth_key_id associated with the provided auth_key
        algorithm (str, optional): The hash algorithm to use with the HMAC authentication scheme (SHA256 | BLAKE2b512 | SHA3-256)

    Returns:
        A new Credentials object.
    """
    def __init__(self, dragonchain_id=None, auth_key=None, auth_key_id=None, algorithm='SHA256'):
        if dragonchain_id is None:
            self.get_dragonchain_id()
        else:
            logger.debug('dragonchain_id provided in constructor parameters, using this excplicitly')
            if isinstance(dragonchain_id, str):
                self.dragonchain_id = dragonchain_id
            else:
                raise TypeError('Parameter "dragonchain_id" must be of type str.')
            logger.debug('Loaded dragonchain id {}'.format(dragonchain_id))

        if auth_key is None or auth_key_id is None:
            self.get_credentials()
        else:
            logger.debug('Auth keys provided in constructor parameters, using these excplicitly')
            if isinstance(auth_key, str) and isinstance(auth_key_id, str):
                self.auth_key = auth_key
                self.auth_key_id = auth_key_id
            else:
                raise TypeError('Parameter "auth_key" and "auth_key_id" must be of type str.')
        logger.debug('Loaded auth key with id: {}'.format(self.auth_key_id))

        self.update_algorithm(algorithm)
        logger.debug('Created credentials client for {}'.format(dragonchain_id))

    def update_algorithm(self, algorithm):
        """Update the algorithm to use with these credentials

        Args:
            algorithm (str): The algorithm to use. Valid options are BLAKE2b512, SHA256, or SHA3-256

        Returns:
            None, sets the HMAC signing algorithm for this credential instance
        """
        if isinstance(algorithm, str):
            self.hash_method = self.get_hash_method(algorithm)
            logger.debug('Valid hash method provided, setting algorithm')
            self.algorithm = algorithm
            logger.info('Updated hashing algorithm to {}'.format(algorithm))
        else:
            raise TypeError('Parameter "algorithm" must be of type str.')

    def get_dragonchain_id(self):
        """Get the dragonchain id if not provided. First checks environment, then configuration files

        Returns:
            None, sets class dragonchain_id instance variable
        """
        logger.debug('Checking if dragonchain_id is in the environment')
        self.dragonchain_id = os.environ.get('DRAGONCHAIN_ID')
        if not self.dragonchain_id:
            logger.debug('dragonchain_id isn\'t in the environment, trying to load default from ini config file')
            try:
                # Check config ini file if ID isn't provided explicitly or in environment
                config = configparser.ConfigParser()
                config.read(Credentials.get_credential_file_path())
                self.dragonchain_id = config.get('default', 'dragonchain_id')
            except (configparser.NoSectionError, configparser.NoOptionError):
                raise exceptions.DragonchainIdentityNotFound('Could not locate dragonchain_id.')

    def get_credentials(self):
        """Get an auth_key/auth_key_id pair if not provided. First checks environment, then configuration files, then smart contract location

        Returns:
            None, sets the class auth_key and auth_key_id instance variables
        """
        if self.get_environment_credentials():
            return
        logger.debug('Credentials aren\'t in environment, trying to load from ini config file')
        if self.get_config_credentials():
            return
        logger.debug('Credentials aren\'t in config file, trying to load as a smart contract')
        if self.get_smart_contract_credentials():
            return
        raise exceptions.DragonchainIdentityNotFound('Unable to locate dragonchain authorization credentials')

    def get_environment_credentials(self):
        """Attempt to get credentials for a dragonchain from the environment

        Returns:
            Boolean True if successful, otherwise False
        """
        logger.debug('Checking if credentials are in the environment')
        self.auth_key = os.environ.get('AUTH_KEY')
        self.auth_key_id = os.environ.get('AUTH_KEY_ID')
        return bool(self.auth_key) and bool(self.auth_key_id)

    def get_config_credentials(self):
        """Attempt to get credentials for a dragonchain from the credentials file

        Returns:
            Boolean True if successful, otherwise False
        """
        try:
            # If both keys aren't in environment variables, check config file
            config = configparser.ConfigParser()
            config.read(Credentials.get_credential_file_path())
            self.auth_key = config.get(self.dragonchain_id, 'auth_key')
            self.auth_key_id = config.get(self.dragonchain_id, 'auth_key_id')
            return True
        except (configparser.NoSectionError, configparser.NoOptionError):
            logger.debug('Could not locate authorization credentials.')
            return False

    def get_smart_contract_credentials(self):
        """Attempt to get credentials for a dragonchain from the standard location for a smart contract

        Returns:
            Boolean True if successful, otherwise False
        """
        try:
            base_path = os.path.join(os.path.abspath(os.sep), 'var', 'openfaas', 'secrets')
            auth_key = open(os.path.join(base_path, 'sc-{}-secret-key'.format(os.environ.get('SMART_CONTRACT_ID'))), 'r').read()
            auth_key_id = open(os.path.join(base_path, 'sc-{}-auth-key-id'.format(os.environ.get('SMART_CONTRACT_ID'))), 'r').read()
            self.auth_key = auth_key
            self.auth_key_id = auth_key_id
            return True
        except (OSError, IOError, FileNotFoundError):
            logger.debug('Could not locate authorization credentials for smart contracts.')
            return False

    def get_hash_method(self, algorithm):
        """Return a hash method that supports the hashlib .new function

        Args:
            algorithm (str): The algorithm to use. Valid options are BLAKE2b512, SHA256, or SHA3-256

        Returns
            Hash method
        """
        if algorithm == 'SHA256':
            return hashlib.sha256
        if sys.version_info >= (3, 6):
            if algorithm == 'BLAKE2b512':
                return hashlib.blake2b
            elif algorithm == 'SHA3-256':
                return hashlib.sha3_256
        raise NotImplementedError('{} is not a supported hash algorithm.'.format(algorithm))

    def bytes_to_b64_str(self, unencoded_bytes):
        """Take a bytes object and output a base64 python string

        Args:
            unencoded_bytes (str): Python bytes object to encode

        Returns:
            String of the base64 encoded bytes
        """
        if not isinstance(unencoded_bytes, bytes):
            raise TypeError('Parameter "unencoded_bytes" must be of type bytes.')
        return base64.b64encode(unencoded_bytes).decode('utf-8')

    def bytes_from_input(self, input_data):
        """Return python bytes from input (either bytes or utf-8 encodable string)

        Args:
            input_data (bytes or str): input to process

        Returns:
            bytes object representation of input
        """
        if isinstance(input_data, str):
            try:
                return input_data.encode('utf-8')
            except Exception:
                raise ValueError('Parameter "input_data" was of type str but not utf-8 encodable.')
        elif isinstance(input_data, bytes):
            return input_data
        else:
            raise TypeError('Parameter "input_data" must be of type str or bytes.')

    def base64_encode_input(self, input_data):
        """Base64 encode some input (either bytes or utf-8 encodable string)

        Args:
            input_data (bytes or str): data to encode

        Returns:
            string of base64 encoded input
        """
        return self.bytes_to_b64_str(self.bytes_from_input(input_data))

    def hash_input(self, input_data):
        """Hash some input_data with a specified (supported) hash type

        Args:
            input_data (str or bytes): data to hash

        Returns:
            bytes of hashed input
        """
        return self.hash_method(self.bytes_from_input(input_data)).digest()

    def create_hmac(self, secret, message):
        """Create an hmac from a given hash type, secret, and message

        Args:
            secret (str): The secret to be used to generate the hmac
            message (bytes or str): The message to use as in the hmac generation

        Returns:
            Bytes for the generated hmac
        """
        return hmac.new(key=self.bytes_from_input(secret), msg=self.bytes_from_input(message), digestmod=self.hash_method).digest()

    def compare_hmac(self, hmac_string, secret, message):
        """Compare a provided base64 encoded hmac string with a generated hmac from the provided secret/message

        Args:
            hmac_string (string): Base64 string of the hmac to compare
            secret (string): The secret to be used to generate the hmac to compare
            message (bytes or string): The message to use with in the hmac generation to compare

        Returns:
            bool if hmac matches or not
        """
        return hmac.compare_digest(base64.b64decode(hmac_string), self.create_hmac(secret, message))

    def hmac_message_string(self, http_verb, path, timestamp, content_type='', content=''):
        """Generate the HMAC message string given the appropriate inputs

        Args:
            http_verb (str): HTTP verb of the request
            path (str): Full path of the request after the FQDN (including any query parameters) (e.g. /transaction)
            timestamp (str): timestamp of the request (must match timestamp header)
            content_type (str): content-type header of the request (if it exists)
            content (bytes or str): byte object of the body of the request (if it exists)

        Returns:
            string to use as the message in HMAC generation
        """
        return '{}\n{}\n{}\n{}\n{}\n{}'.format(http_verb.upper(), path,
                                               self.dragonchain_id, timestamp, content_type,
                                               self.bytes_to_b64_str(self.hash_input(content)))

    def get_authorization(self, http_verb, path, timestamp, content_type='', content=''):
        """Create an authorization header for making requests to a Dragonchain

        Args:
            http_verb (str): HTTP verb of the request
            path (str): full path of the request after the FQDN (including any query parameters) (e.g. /transaction)
            timestamp (str): timestamp of the request (must match timestamp header)
            content_type (str): content-type header of the request (if it exists)
            content (str or bytes): byte object of the body of the request (if it exists)

        Returns:
            String of generated authorization header
        """
        logger.debug('Creating Authorization header for request {} {}'.format(http_verb, path))
        message_string = self.hmac_message_string(http_verb, path, timestamp, content_type, content)
        logger.debug('HMAC message string:\n{}'.format(message_string))
        hmac = self.bytes_to_b64_str(self.create_hmac(self.auth_key, message_string))
        logger.debug('Generated Base64 HMAC string: {}'.format(hmac))
        return 'DC1-HMAC-{} {}:{}'.format(self.algorithm, self.auth_key_id, hmac)

    @staticmethod
    def get_credential_file_path():
        """Get the path for the credential file depending on the OS

        Returns:
            String of the credential file path
        """
        if os.name == 'nt':
            logger.debug('Windows OS detected')
            path = os.path.join(os.path.expandvars('%LOCALAPPDATA%'), 'dragonchain', 'credentials')
        else:
            logger.debug('Posix OS detected')
            path = os.path.join(os.path.expanduser("~"), '.dragonchain', 'credentials')
        logger.debug('Credentials file path: {}'.format(path))
        return path



# Copyright 2019 Dragonchain, Inc. or its affiliates. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dragonchain_sdk.dragonchain_client import Client
import logging

__author__ = 'Dragonchain'
__version__ = '2.0.1'


def set_stream_logger(name='dragonchain_sdk', level=logging.DEBUG, format_string=None):
    """Set a stream logger for a module. You can set name to ``''`` to log everything.

    Args:
        name (str): Name of the module to set the stream logger for
        level (int): Log level, use logging module constants
        format_string (str): Logging format

    Returns:
        None, sets a stream logger
    """
    if format_string is None:
        format_string = '%(asctime)s %(name)s [%(levelname)s] %(message)s'

    logger = logging.getLogger(name)
    logger.setLevel(level)
    handler = logging.StreamHandler()
    handler.setLevel(level)
    formatter = logging.Formatter(format_string)
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def client(*args, **kwargs):
    """Create a client for interacting with a Dragonchain. See ``dragonchain_sdk.dragonchain_client.Client`` for parameters."""
    return Client(*args, **kwargs)


logging.getLogger('dragonchain_sdk').addHandler(logging.NullHandler())



#!/usr/bin/env python3

import os
import re
from setuptools import setup, find_packages


def get_version():
    root = os.path.dirname(os.path.abspath(__file__))
    init = open(os.path.join(root, 'dragonchain_sdk', '__init__.py')).read()
    return re.compile(r'''__version__ = ['"]([0-9.]+)['"]''').search(init).group(1)


setup(
    name='dragonchain_sdk',
    version=get_version(),
    author='Dragonchain',
    author_email='support@dragonchain.com',
    description='Dragonchain SDK for Python',
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    url='https://github.com/dragonchain-inc/dragonchain-sdk-python',
    packages=find_packages(exclude=['*.tests', '*.tests.*', 'tests.*', 'tests']),
    scripts=[],
    install_requires=[
        'requests'
    ],
    license='Apache License 2.0',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3 :: Only'
    ]
)
